# unit tests for init_tools.py

import pytest
import unittest.mock as mock

import numpy as np
import os
from netCDF4 import Dataset

from pyDeltaRCM.model import DeltaModel

from . import utilities


class TestModelDomainSetup:

    def test_inlet_size_specified(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Length', 4000.)
        utilities.write_parameter_to_file(f, 'Width', 8000.)
        utilities.write_parameter_to_file(f, 'dx', 20)
        utilities.write_parameter_to_file(f, 'N0_meters', 150)
        utilities.write_parameter_to_file(f, 'L0_meters', 200)
        f.close()
        delta = DeltaModel(input_file=p)
        assert delta.N0 == 8
        assert delta.L0 == 10

    def test_inlet_size_set_to_one_fourth_domain(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Length', 4000.)
        utilities.write_parameter_to_file(f, 'Width', 8000.)
        utilities.write_parameter_to_file(f, 'dx', 20)
        utilities.write_parameter_to_file(f, 'N0_meters', 5500)
        utilities.write_parameter_to_file(f, 'L0_meters', 3300)
        f.close()
        delta = DeltaModel(input_file=p)
        assert delta.N0 == 100
        assert delta.L0 == 50

    def test_x(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.x[0][-1] == 199

    def test_y(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.y[0][-1] == 0

    def test_cell_type(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        # wall type in corner
        assert delta.cell_type[0, 0] == -2

    def test_eta(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.eta[10, 10] == -5

    def test_stage(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.stage[10, 10] == 0.0

    def test_depth(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.depth[10, 10] == 5

    def test_qx(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        # prescribe the qx at the inlet
        assert delta.qx[0, delta.CTR] == 5

    def test_qy(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.qy[0, delta.CTR] == 0

    def test_qxn(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.qxn[0, 0] == 0

    def test_qyn(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.qyn[0, 0] == 0

    def test_qwn(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.qwn[0, 0] == 0

    def test_ux(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.ux[0, delta.CTR] == 1

    def test_uy(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.uy[0, delta.CTR] == 0

    def test_uw(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.uw[0, delta.CTR] == 1

    def test_qs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.qs[5, 5] == 0

    def test_Vp_dep_sand(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert np.any(delta.Vp_dep_sand) == 0

    def test_Vp_dep_mud(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert np.any(delta.Vp_dep_mud) == 0

    def test_free_surf_flag(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert np.any(delta.free_surf_flag) == 0

    def test_indices(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert np.any(delta.free_surf_walk_inds) == 0

    def test_sfc_visit(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert np.any(delta.sfc_visit) == 0

    def test_sfc_sum(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert np.any(delta.sfc_sum) == 0


class TestCreateBoundaryConditions:

    # base case during init is covered by tests elsewhere!

    def test_change_variable_updated_bcs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # what is Qw0 before?
        assert _delta.Qw0 == 1250.0

        # change u0
        _delta.u0 = 2

        # nothing has happened to Qw0 yet
        assert _delta.u0 == 2
        assert _delta.Qw0 == 1250.0

        # now call to recreate, see what happened
        _delta.create_boundary_conditions()
        assert _delta.u0 == 2
        assert _delta.Qw0 == 2500.0


class TestInitSubsidence:

    def test_subsidence_bounds(self, tmp_path):
        """Test subsidence bounds."""

        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'Length', 600.)
        utilities.write_parameter_to_file(f, 'Width', 600.)
        utilities.write_parameter_to_file(f, 'dx', 5)
        utilities.write_parameter_to_file(f, 'toggle_subsidence', True)
        f.close()

        delta = DeltaModel(input_file=p)
        # assert subsidence mask is binary
        assert np.all(delta.subsidence_mask ==
                      delta.subsidence_mask.astype(bool))
        # check specific regions
        assert np.all(delta.subsidence_mask[delta.L0:, :] == 1)
        assert np.all(delta.subsidence_mask[:delta.L0, :] == 0)


class TestLoadCheckpoint:

    @mock.patch('pyDeltaRCM.shared_tools.set_random_state')
    def test_load_standard_grid(self, patched, tmp_path):
        """Test that a run can be resumed when there are outputs.
        """
        # create one delta, just to have a checkpoint file
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)

        # make mocks
        _delta.log_info = mock.MagicMock()
        _delta.logger = mock.MagicMock()
        _delta.init_output_file = mock.MagicMock()

        # close the file so can be safely opened in load
        _delta.output_netcdf.close()

        # check checkpoint exists
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))

        # now mess up a field
        _eta0 = np.copy(_delta.eta)
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field
        assert np.all(_delta.eta == _rand_field)

        # now resume from the checkpoint to restore the field
        _delta.load_checkpoint()

        # check that fields match
        assert np.all(_delta.eta == _eta0)

        # assertions on function calls
        _call = [mock.call('Renaming old NetCDF4 output file', verbosity=2)]
        _delta.log_info.assert_has_calls(_call, any_order=True)
        _delta.logger.assert_not_called()
        _delta.init_output_file.assert_not_called()
        patched.assert_called()

    @mock.patch('pyDeltaRCM.shared_tools.set_random_state')
    def test_load_wo_netcdf_not_expected(self, patched, tmp_path):
        """
        Test that a checkpoint can be loaded when the load does not expect
        there to be any netcdf file.
        """
        # create one delta, just to have a checkpoint file
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True})
        _delta = DeltaModel(input_file=p)

        # make mocks
        _delta.log_info = mock.MagicMock()
        _delta.logger = mock.MagicMock()
        _delta.init_output_file = mock.MagicMock()

        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))
        assert not os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))

        # now mess up a field
        _eta0 = np.copy(_delta.eta)
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field
        assert np.all(_delta.eta == _rand_field)

        # now resume from the checkpoint to restore the field
        _delta.load_checkpoint()

        # check that fields match
        assert np.all(_delta.eta == _eta0)

        # assertions on function calls
        _delta.log_info.assert_called()
        _delta.logger.assert_not_called()
        _delta.init_output_file.assert_not_called()
        patched.assert_called()

    @mock.patch('pyDeltaRCM.shared_tools.set_random_state')
    def test_load_wo_netcdf_expected(self, patched, tmp_path):
        """
        Test that a checkpoint can be loaded when the load expects there to be
        a netcdf file. This will create a new netcdf file and raise a
        warning.
        """
        # define a yaml with NO outputs, but checkpoint
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)

        # make mocks
        _delta.log_info = mock.MagicMock()
        _delta.logger = mock.MagicMock()
        _delta.init_output_file = mock.MagicMock()

        # close the file so can be safely opened in load
        _delta.output_netcdf.close()

        # check that files exist, and then delete nc
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))
        os.remove(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))

        # now mess up a field
        _eta0 = np.copy(_delta.eta)
        _rand_field = np.random.uniform(0, 1, size=_delta.eta.shape)
        _delta.eta = _rand_field
        assert np.all(_delta.eta == _rand_field)

        # now resume from the checkpoint to restore the field
        with pytest.warns(UserWarning, match=r'NetCDF4 output *.'):
            _delta.load_checkpoint()

        # check that fields match
        assert np.all(_delta.eta == _eta0)
        assert _delta._save_iter == 0

        # assertions on function calls
        _delta.log_info.assert_called()
        _delta.logger.warning.assert_called()
        _delta.init_output_file.assert_called()
        patched.assert_called()

    @mock.patch('pyDeltaRCM.shared_tools.set_random_state')
    def test_load_already_open_netcdf_error(self, patched, tmp_path):
        """
        Test that a checkpoint can be loaded when the load expects there to be
        a netcdf file. This will create a new netcdf file and raise a
        warning.
        """
        # define a yaml with an output, and checkpoint
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)

        # make mocks
        _delta.log_info = mock.MagicMock()
        _delta.logger = mock.MagicMock()
        _delta.init_output_file = mock.MagicMock()

        # close the file so can be safely opened in load
        _delta.output_netcdf.close()

        # check that files exist, and then open the nc back up
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))
        assert os.path.isfile(os.path.join(
            _delta.prefix, 'checkpoint.npz'))
        _ = Dataset(os.path.join(
            _delta.prefix, 'pyDeltaRCM_output.nc'))

        # now try to resume a model and should throw error
        with pytest.raises(RuntimeError):
            _delta.load_checkpoint()


class TestSettingConstants:
    """
    tests for all of the constants
    """

    def test_set_constant_g(self, tmp_path):
        """
        check gravity
        """
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.g == 9.81

    def test_set_constant_distances(self, tmp_path):
        """
        check distances
        """
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.distances[0, 0] == pytest.approx(np.sqrt(2))

    def test_set_ivec(self, tmp_path):
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.ivec[0, 0] == pytest.approx(-np.sqrt(0.5))

    def test_set_jvec(self, tmp_path):
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.jvec[0, 0] == pytest.approx(-np.sqrt(0.5))

    def test_set_iwalk(self, tmp_path):
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.iwalk[0, 0] == -1

    def test_set_jwalk(self, tmp_path):
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.jwalk[0, 0] == -1

    def test_kernel1(self, tmp_path):
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.kernel1[0, 0] == 1

    def test_kernel2(self, tmp_path):
        # create delta from default options
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.kernel2[0, 0] == 1


class TestSettingParametersFromYAMLFile:
    # tests for attrs set during yaml parsing

    def test_init_verbose_default(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)
        assert delta.verbose == 0

    def test_init_verbose(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'verbose': 1})
        delta = DeltaModel(input_file=p)
        assert delta.verbose == 1

    def test_init_seed_zero(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'seed': 0})
        delta = DeltaModel(input_file=p)
        assert delta.seed == 0

    def test_init_Np_water(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'Np_water': 50})
        _delta = DeltaModel(input_file=p)
        assert _delta.init_Np_water == 50

    def test_init_Np_sed(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'Np_sed': 60})
        _delta = DeltaModel(input_file=p)
        assert _delta.init_Np_sed == 60

    def test_dx(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'dx': 20})
        _delta = DeltaModel(input_file=p)
        assert _delta.dx == 20

    def test_itermax(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'itermax': 6})
        _delta = DeltaModel(input_file=p)
        assert _delta.itermax == 6

    def test_h0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'h0': 7.5})
        _delta = DeltaModel(input_file=p)
        assert _delta.h0 == 7.5

        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'h0': int(7)})
        _delta = DeltaModel(input_file=p)
        assert _delta.h0 == 7

    def test_hb(self, tmp_path):
        # take default from h0 if not given:
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'h0': 7.5})
        _delta = DeltaModel(input_file=p)
        assert _delta.h0 == 7.5
        assert _delta.hb == 7.5

        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'hb': 7.5})
        _delta = DeltaModel(input_file=p)
        assert _delta.h0 == 5
        assert _delta.hb == 7.5

        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'hb': int(7)})
        _delta = DeltaModel(input_file=p)
        assert _delta.h0 == 5
        assert _delta.hb == 7

    def test_Nsmooth(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'Nsmooth': 6})
        _delta = DeltaModel(input_file=p)
        assert _delta.Nsmooth == 6

    def test_SLR(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'SLR': 0.01})
        _delta = DeltaModel(input_file=p)
        assert _delta.SLR == 0.01

    def test_omega_flow(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'omega_flow': 0.8})
        _delta = DeltaModel(input_file=p)
        assert _delta.omega_flow == 0.8

    def test_lambda(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'sed_lag': 0.8})
        _delta = DeltaModel(input_file=p)
        assert _delta._lambda == 0.8

    def test_alpha(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'alpha': 0.25})
        _delta = DeltaModel(input_file=p)
        assert _delta.alpha == 0.25

    def test_stepmax_default(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'alpha': 0.25})
        _delta = DeltaModel(input_file=p)
        assert _delta.stepmax == 2 * (_delta.L + _delta.W)

    def test_stepmax_integer(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'stepmax': 10})
        _delta = DeltaModel(input_file=p)
        assert _delta.stepmax == 10

    def test_stepmax_float(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'stepmax': 11.0})
        _delta = DeltaModel(input_file=p)
        assert _delta.stepmax == 11

    def test_save_eta_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert 'eta' in _delta._save_var_list.keys()

    def test_save_depth_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_depth_grids': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert 'depth' in _delta._save_var_list.keys()

    def test_save_stage_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_stage_grids': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert 'stage' in _delta._save_var_list.keys()

    def test_save_discharge_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_discharge_grids': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert 'discharge' in _delta._save_var_list.keys()

    def test_save_velocity_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_velocity_grids': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert 'velocity' in _delta._save_var_list.keys()

    def test_save_sedflux_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_sedflux_grids': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert 'sedflux' in _delta._save_var_list.keys()

    def test_save_sandfrac_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_sandfrac_grids': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert 'sandfrac' in _delta._save_var_list.keys()

    def test_save_discharge_components(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_discharge_components': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert _delta._save_discharge_components is True
        assert 'discharge_x' in _delta._save_var_list.keys()
        assert 'discharge_y' in _delta._save_var_list.keys()

    def test_save_velocity_components(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_velocity_components': True})
        _delta = DeltaModel(input_file=p)
        assert _delta._save_any_grids is True
        assert len(_delta._save_fig_list) == 0
        assert _delta._save_velocity_components is True
        assert 'velocity_x' in _delta._save_var_list.keys()
        assert 'velocity_y' in _delta._save_var_list.keys()

    def test_save_eta_figs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_eta_figs': True})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) > 0
        assert _delta._save_any_grids is False

    def test_save_depth_figs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_depth_figs': True})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) > 0
        assert _delta._save_any_grids is False

    def test_save_stage_figs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_stage_figs': True})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) > 0
        assert _delta._save_any_grids is False

    def test_save_discharge_figs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_discharge_figs': True})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) > 0
        assert _delta._save_any_grids is False

    def test_save_velocity_figs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_velocity_figs': True})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) > 0
        assert _delta._save_any_grids is False

    def test_save_sedflux_figs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_sedflux_figs': True})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) > 0
        assert _delta._save_any_grids is False

    def test_save_sandfrac_figs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_sandfrac_figs': True})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) > 0
        assert _delta._save_any_grids is False

    def test_save_figs_sequential(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_figs_sequential': False})
        _delta = DeltaModel(input_file=p)
        assert len(_delta._save_fig_list) == 0
        assert _delta._save_any_grids is False
        assert _delta._save_figs_sequential is False

    def test_toggle_subsidence(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'toggle_subsidence': True})
        _delta = DeltaModel(input_file=p)
        assert _delta.toggle_subsidence is True

    def test_start_subsidence(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'start_subsidence': 12345})
        _delta = DeltaModel(input_file=p)
        assert _delta.start_subsidence == 12345

    def test_subsidence_rate(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'toggle_subsidence': True,
                                      'subsidence_rate': 1e-9})
        _delta = DeltaModel(input_file=p)
        assert _delta.subsidence_rate == 1e-9
        assert np.any(_delta.sigma > 0)
        assert np.all(_delta.sigma <= (1e-9 * _delta.dt))

    def test_sand_frac_bc(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'sand_frac_bc': -1})
        _delta = DeltaModel(input_file=p)
        assert _delta.sand_frac_bc == -1


class TestSettingOtherParametersFromYAMLSettings:

    def test_theta_sand(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'coeff_theta_sand': 1.4,
                                      'theta_water': 1.2})
        _delta = DeltaModel(input_file=p)
        assert _delta.theta_sand == 1.68

    def test_theta_mud(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'coeff_theta_mud': 0.8,
                                      'theta_water': 1.3})
        _delta = DeltaModel(input_file=p)
        assert _delta.theta_mud == 1.04

    def test_U_dep_mud(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'coeff_U_dep_mud': 0.4325,
                                      'u0': 2.2})
        _delta = DeltaModel(input_file=p)
        assert _delta.U_dep_mud == 0.9515

    def test_U_ero_sand(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'coeff_U_ero_sand': 1.23,
                                      'u0': 2.2})
        _delta = DeltaModel(input_file=p)
        assert _delta.U_ero_sand == 2.706

    def test_U_ero_mud(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'coeff_U_ero_mud': 1.67,
                                      'u0': 2.2})
        _delta = DeltaModel(input_file=p)
        assert _delta.U_ero_mud == 3.674

    def test_L0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'L0_meters': 100,
                                      'Length': 6000,
                                      'dx': 5})
        _delta = DeltaModel(input_file=p)
        assert _delta.L0 == 20

    def test_N0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5})
        _delta = DeltaModel(input_file=p)
        assert _delta.N0 == 100

    def test_L(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'Length': 1600,
                                      'dx': 20})
        _delta = DeltaModel(input_file=p)
        assert _delta.L == 80

    def test_W(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'Width': 1200,
                                      'dx': 20})
        _delta = DeltaModel(input_file=p)
        assert _delta.W == 60

    def test_u_max(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'u0': 2.3})
        _delta = DeltaModel(input_file=p)
        assert _delta.u_max == 4.6   # == 2*u0

    def test_C0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'C0_percent': 10})
        _delta = DeltaModel(input_file=p)
        assert _delta.C0 == 0.1

    def test_dry_depth(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'h0': 0.5})
        _delta = DeltaModel(input_file=p)
        assert _delta.dry_depth == 0.05

    def test_dry_depth_limiter(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'h0': 20})
        _delta = DeltaModel(input_file=p)
        assert _delta.dry_depth == 0.1

    def test_CTR(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'Length': 4000,
                                      'Width': 6000,
                                      'dx': 10})
        _delta = DeltaModel(input_file=p)
        assert _delta.CTR == 299  # 300th index

    def test_CTR_small(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'Length': 40,
                                      'Width': 40,
                                      'dx': 10})
        _delta = DeltaModel(input_file=p)
        # small case, instead of 4/2-1=1 it is 4/2=2
        assert _delta.CTR == 2

    def test_gamma(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'S0': 0.01,
                                      'dx': 10,
                                      'u0': 3})
        _delta = DeltaModel(input_file=p)
        assert _delta.gamma == pytest.approx(0.10900000)

    def test_V0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'h0': 3,
                                      'dx': 15})
        _delta = DeltaModel(input_file=p)
        assert _delta.V0 == 675

    def test_Qw0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5})
        _delta = DeltaModel(input_file=p)
        assert _delta.N0 == 100
        assert _delta.Qw0 == 800

    def test_qw0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'u0': 0.8,
                                      'h0': 3})
        _delta = DeltaModel(input_file=p)
        assert _delta.qw0 == pytest.approx(2.4)

    def test_Qp_water(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5,
                                      'Np_water': 2300})
        _delta = DeltaModel(input_file=p)
        assert _delta.N0 == 100
        assert _delta.Qw0 == 800
        assert _delta.Qp_water == pytest.approx(0.347826087)

    def test_dVs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5})
        _delta = DeltaModel(input_file=p)
        assert _delta.V0 == 50
        assert _delta.N0 == 100
        assert _delta.Qw0 == 800
        assert _delta.dVs == 50000

    def test_Qs0(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'C0_percent': 10,
                                      'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5})
        _delta = DeltaModel(input_file=p)
        assert _delta.Qw0 == 800
        assert _delta.Qs0 == pytest.approx(800 * 0.1)

    def test_Vp_sed(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5,
                                      'Np_sed': 1450})
        _delta = DeltaModel(input_file=p)
        assert _delta.dVs == 50000
        assert _delta.Vp_sed == 50000 / 1450

    def test_stepmax_and_size_indices(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'Length': 1600,
                                      'Width': 1200,
                                      'dx': 20})
        _delta = DeltaModel(input_file=p)
        assert _delta.L == 80
        assert _delta.W == 60
        assert _delta.stepmax == (80 + 60) * 2
        assert _delta.size_indices == (80 + 60)

    def test_dt(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'C0_percent': 10,
                                      'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5})
        _delta = DeltaModel(input_file=p)
        assert _delta.Qw0 == 800
        assert _delta.dVs == 50000
        assert _delta.Qs0 == pytest.approx(800 * 0.1)
        assert _delta.dt == 625

    def test_omega_flow_iter(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'itermax': 7})
        _delta = DeltaModel(input_file=p)
        assert _delta.omega_flow_iter == pytest.approx(2 / 7)

    def test_N_crossdiff(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5})
        _delta = DeltaModel(input_file=p)
        assert _delta.V0 == 50
        assert _delta.N0 == 100
        assert _delta.Qw0 == 800
        assert _delta.dVs == 50000
        assert _delta.N_crossdiff == int(round(50000 / 50))

    def test_diffusion_multiplier(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'u0': 0.8,
                                      'h0': 2,
                                      'N0_meters': 500,
                                      'Width': 6000,
                                      'dx': 5,
                                      'alpha': 0.3,
                                      'C0_percent': 10})
        _delta = DeltaModel(input_file=p)
        assert _delta.V0 == 50
        assert _delta.N0 == 100
        assert _delta.Qw0 == 800
        assert _delta.dVs == 50000
        assert _delta.Qs0 == pytest.approx(800 * 0.1)
        assert _delta.dt == 625
        assert _delta.N_crossdiff == int(round(50000 / 50))
        assert _delta.diffusion_multiplier == (625 / 1000 * 0.3 * 0.5 / 5**2)

    def test_active_layer_thickness_float(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'active_layer_thickness': 2.7})
        _delta = DeltaModel(input_file=p)
        assert _delta.active_layer_thickness == 2.7

    def test_active_layer_thickness_int(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'active_layer_thickness': 2})
        _delta = DeltaModel(input_file=p)
        assert _delta.active_layer_thickness == 2

    def test_active_layer_thickness_default(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)
        assert _delta.active_layer_thickness == _delta.h0 / 2


class TestInitMetadataList:

    def test_save_list_exists(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        f.close()
        delta = DeltaModel(input_file=p)
        # check things about the metadata
        assert hasattr(delta, '_save_var_list')
        assert type(delta._save_var_list) == dict
        assert 'meta' in delta._save_var_list.keys()
        # save meta not on, so check that it is empty
        assert delta._save_var_list['meta'] == {}

    def test_default_meta_list(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'save_metadata', True)
        utilities.write_parameter_to_file(f, 'legacy_netcdf', False)
        f.close()
        delta = DeltaModel(input_file=p)
        # check things about the metadata
        assert hasattr(delta, '_save_var_list')
        assert type(delta._save_var_list) == dict
        assert 'meta' in delta._save_var_list.keys()
        # save meta on, so check that some expected values are there
        assert 'L0' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['L0'] == ['L0', 'cells', 'i8', ()]
        assert 'H_SL' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['H_SL'] == \
            [None, 'meters', 'f4', 'time']

    def test_default_meta_list_legacy(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'save_metadata', True)
        utilities.write_parameter_to_file(f, 'legacy_netcdf', True)
        f.close()
        delta = DeltaModel(input_file=p)
        # check things about the metadata
        assert hasattr(delta, '_save_var_list')
        assert type(delta._save_var_list) == dict
        assert 'meta' in delta._save_var_list.keys()
        # save meta on, so check that some expected values are there
        assert 'L0' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['L0'] == ['L0', 'cells', 'i8', ()]
        assert 'H_SL' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['H_SL'] == \
            [None, 'meters', 'f4', 'total_time']

    def test_netcdf_vars(self, tmp_path):
        # test that stuff makes it to the netcdf file as expected
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(f, 'save_metadata', True)
        utilities.write_parameter_to_file(f, 'legacy_netcdf', False)
        f.close()
        delta = DeltaModel(input_file=p)
        # check things about the metadata
        assert hasattr(delta, '_save_var_list')
        assert type(delta._save_var_list) == dict
        assert 'meta' in delta._save_var_list.keys()
        # save meta on, so check that some expected values are there
        assert 'L0' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['L0'] == ['L0', 'cells', 'i8', ()]
        assert 'H_SL' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['H_SL'] == \
            [None, 'meters', 'f4', 'time']
        # check save var list for eta
        assert 'eta' in delta._save_var_list.keys()
        assert delta._save_var_list['eta'] == \
            ['eta', 'meters', 'f4', ('time', 'x', 'y')]
        # force save to netcdf
        delta.save_grids_and_figs()
        # close netcdf
        delta.output_netcdf.close()
        # check out the netcdf
        data = Dataset(os.path.join(delta.prefix, 'pyDeltaRCM_output.nc'),
                       'r+', format='NETCDF4')
        # check for meta group
        assert 'meta' in data.groups
        # check for L0 a single value metadata
        assert 'L0' in data['meta'].variables
        assert data['meta']['L0'][0].data == delta.L0
        # check H_SL a vector of metadata
        assert 'H_SL' in data['meta'].variables
        assert data['meta']['H_SL'].dimensions == ('time',)
        assert data['time'].shape == data['meta']['H_SL'].shape
        # check on the eta grid
        assert 'eta' in data.variables
        assert data['eta'].shape[0] == data['time'].shape[0]
        assert data['eta'].shape[1] == delta.L
        assert data['eta'].shape[2] == delta.W

    def test_netcdf_vars_legacy(self, tmp_path):
        # test that stuff makes it to the netcdf file as expected
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        utilities.write_parameter_to_file(f, 'save_metadata', True)
        utilities.write_parameter_to_file(f, 'legacy_netcdf', True)
        f.close()
        delta = DeltaModel(input_file=p)
        # check things about the metadata
        assert hasattr(delta, '_save_var_list')
        assert type(delta._save_var_list) == dict
        assert 'meta' in delta._save_var_list.keys()
        # save meta on, so check that some expected values are there
        assert 'L0' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['L0'] == ['L0', 'cells', 'i8', ()]
        assert 'H_SL' in delta._save_var_list['meta'].keys()
        assert delta._save_var_list['meta']['H_SL'] == \
            [None, 'meters', 'f4', 'total_time']
        # check save var list for eta
        assert 'eta' in delta._save_var_list.keys()
        assert delta._save_var_list['eta'] == \
            ['eta', 'meters', 'f4', ('total_time', 'length', 'width')]
        # force save to netcdf
        delta.save_grids_and_figs()
        # close netcdf
        delta.output_netcdf.close()
        # check out the netcdf
        data = Dataset(os.path.join(delta.prefix, 'pyDeltaRCM_output.nc'),
                       'r+', format='NETCDF4')
        # check for meta group
        assert 'meta' in data.groups
        # check for L0 a single value metadata
        assert 'L0' in data['meta'].variables
        assert data['meta']['L0'][0].data == delta.L0
        # check H_SL a vector of metadata
        assert 'H_SL' in data['meta'].variables
        assert data['meta']['H_SL'].dimensions == ('total_time',)
        assert data['time'].shape == data['meta']['H_SL'].shape
        # check on the eta grid
        assert 'eta' in data.variables
        assert data['eta'].shape[0] == data['time'].shape[0]
        assert data['eta'].shape[1] == delta.L
        assert data['eta'].shape[2] == delta.W




import numpy as np
import yaml
import re
import contextlib
import tempfile
import os

from numba import njit, _helperlib

# tools shared between deltaRCM water and sediment routing

# useful constants
sec_in_day = 86400
day_in_yr = 365.25
earth_grav = 9.81


@njit
def set_random_seed(_seed):
    """Set the random seed from an integer.

    Set the seed of the random number generator with a single integer.
    Importantly, this function will affect the state of the `numba` random
    number generator, rather than the `numpy` generator.

    This function is called during model intialization, but is **not** called
    during loading a model from checkpoint
    (:meth:`~pyDeltaRCM.init_tools.init_tools.load_checkpoint`) where the
    random number generator state is configured via the
    :func:`set_random_state` function.

    .. important::

        This function is preferred to setting the random seed directly. If you
        are working with model subclassing, interacting with the `numpy`
        random number generator directly will lead to a model that is **not
        reproducible!**

    Internally, the method is simply the below line; however it is *crucial*
    this happens inside a jitted function, in order to affect the state of the
    `numba` random number generator seed.

    .. code::

        np.random.seed(_seed)

    Parameters
    ----------
    _seed : :obj:`int`
        Integer to use as the random number generator seed. Should be in
        interval ``[0, 2^32]``.
    """
    np.random.seed(_seed)


def get_random_state():
    """Get the random state as a tuple.

    Get the random state from a tuple in the form returned by
    :func:`numba._helperlib.rnd_get_state`. This tuple contains the
    necessary information for resuming a checkpoint from the exact same random
    number generator state.

    .. important::

        You probably do not need to use this function. Be sure you know
        *exactly* how you are affecting the random number generator state
        before using this function, or you are likely to create a model that
        cannot be peproduced.

    See also :func:`set_random_seed`.

    Returns
    -------
    state : :obj:`tuple`
        Random number generator state as a tuple.
    """
    ptr = _helperlib.rnd_get_np_state_ptr()
    return _helperlib.rnd_get_state(ptr)


def set_random_state(_state_tuple):
    """Set the random state from a tuple.

    Set the random state from a tuple in the form returned by
    :func:`numba._helperlib.rnd_get_state`.

    .. important::

        You probably do not need to use this function. Be sure you know
        *exactly* how you are affecting the random number generator state
        before using this function, or you are likely to create a model that
        cannot be peproduced.

    See also :func:`set_random_seed`.

    Parameters
    ----------
    state : :obj:`tuple`
        Random number generator state as a tuple.
    """
    ptr = _helperlib.rnd_get_np_state_ptr()
    _helperlib.rnd_set_state(ptr, _state_tuple)


@njit
def get_random_uniform(limit):
    """Get a random number from the uniform distribution.

    Get a random number from the uniform distribution, defined over the
    interval ``[0, limit]``, where `limit` is an input parameter (usually
    ``1``).

    .. hint::

        Function returns only a single float. Wrap the call in a simple for
        loop to get many random numbers.

    Parameters
    ----------
    limit : :obj:`float`
        Upper limit to distribution to draw from.

    Returns
    -------
    :obj:`float`
        Float value from the random number generator in the interval defined.

    Examples
    --------
    .. testsetup::

        from pyDeltaRCM.shared_tools import get_random_uniform, set_random_seed
        set_random_seed(0)

    .. doctest::

        >>> get_random_uniform(1)
        0.5488135039273248

        >>> get_random_uniform(0.1)
        0.07151893663724194

    """
    return np.random.uniform(0, limit)


def get_inlet_weights(inlet):
    """Get weight for inlet parcels.

    This method is called at the top of water and sediment routing steps by
    default, as implemented in the methods
    :obj:`~pyDeltaRCM.water_tools.water_tools.get_inlet_weights_water` and
    :obj:`~pyDeltaRCM.water_tools.water_tools.get_inlet_weights_sediment`.

    The function determines the *weights* describing which inlet cells parcels
    are fed into the domain. As implemented by default, returns "ones" for
    all inlet cell locations, so that the input sediment source is balanced
    across all inlet cells. The returned weight array is typically passed
    to :obj:`get_start_indices` to determine the locations to feed parcels
    into the domain based on the weights determined by this function.

    Parameters
    ----------
    inlet : :obj:`ndarray` of int
        Integer array specifying the inlet cell indices.
    """
    inlet_weights = np.ones_like(inlet)
    return inlet_weights


@njit
def get_start_indices(inlet, inlet_weights, num_starts):
    """Get start indices.

    Reutrn a randomly generated list of starting points for parcel routing.
    These starting points are selected from the `inlet` array.

    Parameters
    ----------
    inlet : ndarray
        Array of inlet cells.

    inlet_weights : ndarray
        Array of weights to select items from :obj:`inlet`. Should have same
        dimensions.

    num_starts : int
        Number of starting points to generate.

    Returns
    -------
    start_indices : ndarray
        :obj:`num_starts` starting points, generated from :obj:`inlet`
         according to weights in :obj:`inlet_weights`. 
    """
    norm_weights = inlet_weights / np.sum(inlet_weights)
    idxs = []
    for _ in np.arange(num_starts):
        idxs.append(random_pick(norm_weights))
    idxs = np.array(idxs)
    return inlet.take(idxs)


@njit
def get_steps(new_direction, iwalk, jwalk):
    """Find the values given the next step.

    Get the steps for updating discharge and velocity arrays based on the
    direction of each step.
    """
    istep = iwalk[new_direction]
    jstep = jwalk[new_direction]
    dist = np.sqrt(istep * istep + jstep * jstep)
    astep = (dist != 0)

    return dist, istep, jstep, astep


@njit
def random_pick(prob):
    """Pick number from weighted array.

    Randomly pick a number weighted by array probabilities (len 9)
    Return the index of the selected weight in array probs
    Takes a numpy array that is the precalculated probability
    around the cell flattened to 1D.

    Examples
    --------

    .. todo:: Examples needed (pull from tests?).

    """
    arr = np.arange(len(prob))
    cumprob = np.cumsum(prob)
    return arr[np.searchsorted(cumprob, get_random_uniform(cumprob[-1]))]


@njit('UniTuple(int64, 2)(int64, UniTuple(int64, 2))')
def custom_unravel(i, shape):
    """Unravel indexes for 2D array.

    This is a jitted function, equivalent to the `numpy` implementation of
    `unravel_index`, but configured to accept only a single integer as
    the index to unravel.

    Parameters
    ----------
    i : :obj:`int`
        Integer index to unravel.

    shape : :obj:`tuple` of :obj:`int`
        Two element tuple of integers indicating the shape of the
        two-dimensional array from which :obj:`i` will be unravelled.

    Returns
    -------
    x : :obj:`int`
        `x` index of :obj:`i` into array of shape :obj:`shape`.

    y : :obj:`int`
        `y` index of :obj:`i` into array of shape :obj:`shape`.

    Examples
    --------
    .. testsetup::

        from pyDeltaRCM.shared_tools import custom_unravel

    .. doctest::

        >>> _shape = (100, 200)  # e.g., delta.eta.shape

        >>> np.unravel_index(5124, _shape)
        (25, 124)

        >>> custom_unravel(5124, _shape)
        (25, 124)
    """
    if i > (shape[1] * shape[0]):
        raise IndexError("Index is out of array bounds")
    x = i // shape[1]
    y = i % shape[1]
    return x, y


@njit('int64(UniTuple(int64, 2), UniTuple(int64, 2))')
def custom_ravel(tup, shape):
    """Ravel indexes for 2D array.

    This is a jitted function, equivalent to the `numpy` implementation of
    `ravel_multi_index`, but configured to accept only a single tuple as
    the index to ravel, and only works for two-dimensional array.

    Parameters
    ----------
    tup : :obj:`tuple` of :obj:`int`
        Two element `tuple` with the `x` and `y` index into an array of
        shape :obj:`shape` to ravel.

    shape : :obj:`tuple` of :obj:`int`
        Two element tuple of integers indicating the shape of the
        two-dimensional array from which :obj:`tup` will be ravelled.

    Returns
    -------
    i : :obj:`int`
        Ravelled integer index into array of shape :obj:`shape`.

    Examples
    --------
    .. testsetup::

        from pyDeltaRCM.shared_tools import custom_ravel

    .. doctest::

        >>> _shape = (100, 200)  # e.g., delta.eta.shape

        >>> np.ravel_multi_index((25, 124), _shape)
        5124

        >>> custom_ravel((25, 124), _shape)
        5124
    """
    if tup[0] > shape[0] or tup[1] > shape[1]:
        raise IndexError("Index is out of array bounds")
    x = tup[0] * shape[1]
    y = tup[1]
    return x + y


@njit
def custom_pad(arr):
    """Pad an array.

    This is a jitted function, equivalent to the `numpy` implementation of
    `pad` with certain optional parameters:

    .. code::

        np.pad(arr, 1, 'edge')

    In pyDeltaRCM model fields (e.g., `depth`) are frequently padded to enable
    straightforward slicing of the neighbors of a given cell; i.e., padding
    guarantees that all cells will have 9 neighbors.

    Parameters
    ----------
    arr : :obj:`ndarray`
        Array to pad.

    Returns
    -------
    pad : :obj:`ndarray`
        Padded array.

    Examples
    --------
    .. testsetup::

        from pyDeltaRCM.shared_tools import custom_pad

    Consider a model domain of size `(4, 8)`

    .. doctest::

        >>> arr = np.arange(32).reshape(4, 8)

        >>> np_pad = np.pad(arr, 1, 'edge')
        >>> cust_pad = custom_pad(arr)
        >>> np.all(np_pad == cust_pad)
        True
        >>> cust_pad.shape
        (6, 10)

    which enables all elements of the original `(4, 8)` array to be safely
    sliced:

    .. doctest::

        >>> for i in range(4):
        ...     for j in range(8):
        ...         slc = cust_pad[i:i+3, j:j+3]

        >>> slc.shape  # will always be a (3, 3) 9-cell neighborhood
        (3, 3)

    """
    old_shape = arr.shape
    new_shape = (old_shape[0]+2, old_shape[1]+2)
    pad = np.zeros(new_shape, dtype=arr.dtype)

    # center
    pad[1:-1, 1:-1] = arr

    # edges
    pad[1:-1, 0] = arr[:, 0]  # left
    pad[1:-1, -1] = arr[:, -1]  # right
    pad[0, 1:-1] = arr[0, :]  # top
    pad[-1, 1:-1] = arr[-1, :]  # bottom

    # corners
    pad[0, 0] = arr[0, 0]  # ul
    pad[0, -1] = arr[0, -1]  # ur
    pad[-1, 0] = arr[-1, 0]  # ll
    pad[-1, -1] = arr[-1, -1]  # lr

    return pad


@njit
def get_weight_sfc_int(stage, stage_nbrs, qx, qy, ivec, jvec, distances):
    """Determine random walk weight surfaces.

    Determines the surfaces for weighting the random walks based on the stage
    and discharge fields.

    .. todo::

        Expand description. Where is it used? Include an example? Equations?
        Link to Hydrodyanmics doc?
    """
    weight_sfc = np.maximum(0, (stage - stage_nbrs) / distances)
    weight_int = np.maximum(0, (qx * jvec + qy * ivec) / distances)
    return weight_sfc, weight_int


def _get_version():
    """Extract version from file.

    Extract version number from single file, and make it availabe everywhere.
    This function is used to make sure that when the version number is
    incremented in the `pyDeltaRCM._version` module, it is properly updated
    throughout documentation and version-controlled release.

    .. note:: You probably don't need to use this function.

    Returns
    -------
    ver_str : :obj:`str`
        Version number, as a string.

    Examples
    --------
    .. doctest::

        >>> pyDeltaRCM.shared_tools._get_version()  # doctest: +SKIP
    """
    from . import _version
    return _version.__version__()


@contextlib.contextmanager
def _docs_temp_directory():
    """Helper for creating and tearing down models in documentation.

    This function should be used as a context manager, to create a DeltaModel
    with a temporary folder as output, rather than anywhere in the project
    structure.

    Examples
    --------
    .. testsetup::

        from pyDeltaRCM.shared_tools import _docs_temp_directory

    .. doctest::

        >>> with _docs_temp_directory() as output_dir:
        ...     delta = pyDeltaRCM.DeltaModel(out_dir=output_dir)
    """
    tmpdir = tempfile.TemporaryDirectory()
    output_path = os.path.join(tmpdir.name, 'output')
    yield output_path
    tmpdir.cleanup()


def custom_yaml_loader():
    """A custom YAML loader to handle scientific notation.

    We are waiting for upstream fix here:
        https://github.com/yaml/pyyaml/pull/174

    Returns
    -------
    loader : :obj:`pyyaml.Loader` with custom resolver

    Examples
    --------
    The custom loader can be used as:

    .. code::

        >>> loader = pyDeltaRCM.shared_tools.custom_yaml_loader()

        >>> a_file = open('/path/to/file.yaml', mode='r')
        >>> yaml_as_dict = yaml.load(a_file, Loader=loader)
    """
    loader = yaml.SafeLoader
    loader.add_implicit_resolver(
        u'tag:yaml.org,2002:float',
        re.compile(r'''^(?:[-+]?(?:[0-9][0-9_]*)\.[0-9_]*(?:[eE][-+]?[0-9]+)?
                       |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
                       |\.[0-9_]+(?:[eE][-+]?[0-9]+)?
                       |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*
                       |[-+]?\.(?:inf|Inf|INF)
                       |\.(?:nan|NaN|NAN))$''', re.X),
        list(u'-+0123456789.'))
    return loader


def scale_model_time(time, If=1, units='seconds'):
    """Scale the model time to "real" time.

    Model time is executed as assumed flooding conditions, and executed at the
    per-second level, with a multi-second timestep. This model design
    implicitly assumes that the delta is *always* receiving a large volume of
    sediment and water at the inlet. This is unrealistic, given that rivers
    flood only during a small portion of the year, and this is when
    morphodynamic activity is largest. See :doc:`../../info/modeltime` for a
    complete description of this assumption, and how to work with the
    assumption in configuring the model.

    Using this assumption, it is possible to scale up model time to "real"
    time, by assuming an *intermittency factor*. This intermittency factor is
    the fraction of unit-time that a river is assumed to be flooding.

    .. math::

        t_r = \dfrac{t}{I_f \cdot S_f}

    where :math:`t` is the model time (:obj:`~pyDeltaRCM.DeltaModel.time`),
    :math:`t_r` is the "real" scaled time, :math:`I_f` is the
    intermittency factor, and :math:`S_f` is the scale factor to convert base
    units of seconds to units specified as an input argument. Note that this
    function uses :obj:`_scale_factor` internally for this conversion.

    Parameters
    ----------
    time : :obj:`float`
        The model time, in seconds.

    If : :obj:`float`, optional
        Intermittency factor, fraction of time represented by morphodynamic
        activity. Should be in interval (0, 1]. Defaults to 1 if not provided,
        i.e., no scaling is performed.

    units : :obj:`str`, optional
        The units to convert the scaled time to. Default is to return the
        scaled time in seconds (`seconds`), but optionally supply argument
        `days` or `years` for unit conversion.

    Returns
    -------
    scaled : :obj:`float`
        Scaled time, in :obj:`units`, assuming the intermittency factor
        :obj:`If`.

    Raises
    ------
    ValueError
        if the value for intermittency is not ``0 < If <= 1``.
    """
    if (If <= 0) or (If > 1):
        raise ValueError(
            'Intermittency `If` is not 0 < If <= 1: %s' % str(If))

    return time / _scale_factor(If, units)


def _scale_factor(If, units):
    """Scaling factor between model time and "real" time.

    The scaling factor relates the model time to a real worl time, by the
    assumed intermittency factor and the user-specified units for output.

    Parameters
    ----------
    If : :obj:`float`
        Intermittency factor, fraction of time represented by morphodynamic
        activity. **Must** be in interval (0, 1].

    units : :obj:`str`
        The units to convert the scaled time to. Must be a string in
        `['seconds', 'days', 'years']`.

    """
    if units == 'seconds':
        S_f = 1
    elif units == 'days':
        S_f = sec_in_day
    elif units == 'years':
        S_f = sec_in_day * day_in_yr
    else:
        raise ValueError('Bad value for `units`: %s' % str(units))
    return (If * S_f)



#! /usr/bin/env python
from setuptools import setup

from pyDeltaRCM import shared_tools

setup(name='pyDeltaRCM',
      version=shared_tools._get_version(),
      author='The DeltaRCM Team',
      license='MIT',
      description="Python version of original Matlab DeltaRCM",
      long_description=open('README.rst').read(),
      packages=['pyDeltaRCM'],
      include_package_data=True,
      url='https://github.com/DeltaRCM/pyDeltaRCM',
      project_urls={
            'Documentation': 'https://deltarcm.org/pyDeltaRCM',
            'Source': 'https://github.com/DeltaRCM/pyDeltaRCM',
            'Tracker': 'https://github.com/deltaRCM/pyDeltaRCM/issues'},
      install_requires=['matplotlib', 'netCDF4', 'scipy', 'numpy', 'pyyaml',
                        'numba'],
      entry_points={
            'console_scripts': [
                  'pyDeltaRCM=pyDeltaRCM.preprocessor:preprocessor_wrapper'],
      }
      )



# unit tests for iteration_tools.py

import pytest

import os
import netCDF4
import numpy as np
import matplotlib.pyplot as plt

import unittest.mock as mock

from pyDeltaRCM.model import DeltaModel
from . import utilities


class TestSolveWaterAndSedimentTimestep:

    def test_solve_water_and_sediment_timestep_defaults(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        # mock top-level methods, verify call was made to each
        delta.log_info = mock.MagicMock()
        delta.init_water_iteration = mock.MagicMock()
        delta.run_water_iteration = mock.MagicMock()
        delta.compute_free_surface = mock.MagicMock()
        delta.finalize_water_iteration = mock.MagicMock()
        delta.route_sediment = mock.MagicMock()

        # run the timestep
        delta.solve_water_and_sediment_timestep()

        # assert that methods are called
        assert delta.init_water_iteration.called is True
        assert delta.run_water_iteration.called is True
        assert delta.compute_free_surface.called is True
        assert delta.finalize_water_iteration.called is True
        _calls = [mock.call(0), mock.call(1), mock.call(2)]
        delta.finalize_water_iteration.assert_has_calls(
            _calls, any_order=False)
        assert delta.finalize_water_iteration.call_count == 3
        assert (delta.route_sediment.called is True)
        assert (delta._is_finalized is False)

    def test_solve_water_and_sediment_timestep_itermax_10(self, tmp_path):
        # create a delta with different itermax
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'itermax': 10})
        delta = DeltaModel(input_file=p)

        # mock top-level methods, verify call was made to each
        delta.log_info = mock.MagicMock()
        delta.init_water_iteration = mock.MagicMock()
        delta.run_water_iteration = mock.MagicMock()
        delta.compute_free_surface = mock.MagicMock()
        delta.finalize_water_iteration = mock.MagicMock()
        delta.route_sediment = mock.MagicMock()

        # run the timestep
        delta.solve_water_and_sediment_timestep()

        # assert that methods are called
        assert delta.init_water_iteration.called is True
        assert delta.run_water_iteration.called is True
        assert delta.compute_free_surface.called is True
        assert delta.finalize_water_iteration.called is True
        _calls = [mock.call(i) for i in range(10)]
        delta.finalize_water_iteration.assert_has_calls(
            _calls, any_order=False)
        assert delta.finalize_water_iteration.call_count == 10
        assert (delta.route_sediment.called is True)
        assert (delta._is_finalized is False)

    def test_run_one_timestep_deprecated(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock top-level methods
        _delta.logger = mock.MagicMock()
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()

        # check warning raised
        with pytest.warns(UserWarning):
            _delta.run_one_timestep()

        # and logged
        assert (_delta.logger.warning.called is True)


class TestFinalizeTimestep:

    def test_finalize_timestep(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'SLR': 0.001})
        delta = DeltaModel(input_file=p)

        # mock the flooding correction and log
        delta.log_info = mock.MagicMock()
        delta.flooding_correction = mock.MagicMock()

        # run the step
        delta.finalize_timestep()

        # assert submethod called once
        delta.flooding_correction.call_count == 1

        # check that sea level rose as expected
        assert delta.H_SL == 25


class TestApplyingSubsidence:

    def test_subsidence_in_update(self, tmp_path):
        # create a delta with subsidence parameters
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'toggle_subsidence': True,
                                      'subsidence_rate': 1e-8,
                                      'start_subsidence': 0,
                                      'save_eta_grids': True,
                                      'seed': 0})
        _delta = DeltaModel(input_file=p)

        # mock the timestep computations
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()

        assert _delta.dt == 25000
        assert _delta.subsidence_rate == 1e-8
        assert np.all(_delta.sigma[:_delta.L0, :] == 0.0)  # outside the sigma mask
        assert np.all(_delta.sigma[_delta.L0:, :] == 0.00025)  # inside the sigma mask
        assert np.all(_delta.eta[_delta.L0:, :] == -_delta.h0)

        _delta.update()
        assert np.all(_delta.eta[_delta.L0-1, :25] == 0.0)
        assert np.all(_delta.eta[_delta.L0:, :] ==
                      pytest.approx(-_delta.h0 - 0.00025))
        _delta.output_netcdf.close()

    def test_subsidence_in_update_delayed_start(self, tmp_path):
        # create a delta with subsidence parameters
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'toggle_subsidence': True,
                                      'subsidence_rate': 1e-8,
                                      'start_subsidence': 25000,
                                      'save_eta_grids': True,
                                      'seed': 0})
        _delta = DeltaModel(input_file=p)

        # mock the timestep computations
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()

        assert _delta.dt == 25000
        assert _delta.subsidence_rate == 1e-8
        assert np.all(_delta.sigma[:_delta.L0, :] == 0.0)  # outside the sigma mask
        assert np.all(_delta.sigma[_delta.L0:, :] == 0.00025)  # inside the sigma mask
        assert np.all(_delta.eta[_delta.L0:, :] == -_delta.h0)

        _delta.update()  # no subsidence applied
        assert _delta.time == 25000
        assert np.all(_delta.eta[_delta.L0-1, :25] == 0.0)
        assert np.all(_delta.eta[_delta.L0:, :] == pytest.approx(-_delta.h0))

        _delta.update()
        assert _delta.time == 50000
        assert np.all(_delta.eta[_delta.L0-1, :25] == 0.0)
        assert np.all(_delta.eta[_delta.L0:, :] ==
                      pytest.approx(-_delta.h0 - 0.00025))
        _delta.output_netcdf.close()

        _delta.solve_water_and_sediment_timestep.call_count == 2

    def test_subsidence_changed_with_timestep(self, tmp_path):
        # create a delta with subsidence parameters
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'toggle_subsidence': True,
                                      'save_eta_grids': True,
                                      'subsidence_rate': 1e-8})
        _delta = DeltaModel(input_file=p)

        assert _delta.dt == 25000
        assert np.all(_delta.sigma[_delta.L0:, :] == 0.00025)
        # use the model setter to adjust the timestep
        _delta.time_step = 86400
        assert np.all(_delta.sigma[_delta.L0:, :] == 0.000864)
        _delta.output_netcdf.close()


class TestOutputCheckpoint:

    def test_save_a_checkpoint_checkpoint_False(self, tmp_path):
        # create a delta with subsidence parameters
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': False})
        _delta = DeltaModel(input_file=p)

        # mock the actual save checkpoint function to see if it was called
        _delta.save_the_checkpoint = mock.MagicMock()
        _delta.log_info = mock.MagicMock()

        _delta.output_checkpoint()

        # assertions
        assert (_delta.save_the_checkpoint.called is False)
        assert (_delta.log_info.called is False)

    def test_save_a_checkpoint_checkpoint_true(self, tmp_path):
        # create a delta with subsidence parameters
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True})
        _delta = DeltaModel(input_file=p)

        # force the time to be greater than the checkpoint interval
        _delta._save_time_since_checkpoint = 2 * _delta._checkpoint_dt

        # mock the actual save checkpoint function to see if it was called
        _delta.save_the_checkpoint = mock.MagicMock()
        _delta.log_info = mock.MagicMock()

        # run the output checkpoint func
        _delta.output_checkpoint()

        # assertions
        assert (_delta.save_the_checkpoint.call_count == 1)
        assert (_delta.log_info.call_count == 1)

    def test_save_a_checkpoint_checkpoint_true_timewarning(self, tmp_path):
        # create a delta with subsidence parameters
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True,
                                      'checkpoint_dt': 864000000})
        _delta = DeltaModel(input_file=p)

        # force the time to be greater than the checkpoint interval
        _delta._save_time_since_checkpoint = 2 * _delta._checkpoint_dt

        # mock the actual save checkpoint function to see if it was called
        _delta.save_the_checkpoint = mock.MagicMock()

        # this warning is only written to log, so mock the logger
        _delta.logger = mock.MagicMock()

        # run the output checkpoint func
        _delta.output_checkpoint()

        # assertions
        assert (_delta.save_the_checkpoint.call_count == 1)
        assert (_delta.logger.warning.call_count == 1)


class TestSaveGridsAndFigs:

    def test_save_no_figs_no_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # mock the actual output routines
        _delta.make_figure = mock.MagicMock()
        _delta.save_figure = mock.MagicMock()
        _delta.save_grids = mock.MagicMock()

        # check nothing created at the start
        _delta.make_figure.call_count == 0
        _delta.save_figure.call_count == 0
        _delta.save_grids.call_count == 0

        # update the delta a few times
        for _t in range(0, 4):
            _delta._time = (_t * _delta._dt)
            _delta.save_grids_and_figs()

        # check nothing after a number of iterations, greater than dt
        assert _delta._time > _delta.save_dt
        _delta.make_figure.call_count == 0
        _delta.save_figure.call_count == 0
        _delta.save_grids.call_count == 0

    def test_save_one_fig_no_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_eta_figs': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # mock the actual output routines
        _delta.make_figure = mock.MagicMock()
        _delta.save_figure = mock.MagicMock()
        _delta.save_grids = mock.MagicMock()

        # check nothing created at the start
        _delta.make_figure.call_count == 0
        _delta.save_figure.call_count == 0
        _delta.save_grids.call_count == 0

        assert (len(_delta._save_fig_list) > 0)
        assert (_delta._save_eta_figs is True)

        # update the delta a few times
        for _t in range(0, 5):
            _delta.save_grids_and_figs()
            _delta._save_iter += 1

        _delta.make_figure.call_count == 5
        _delta.save_figure.call_count == 5
        _delta.save_grids.call_count == 0

    def test_save_one_fig_one_grid(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_eta_grids': True,
                                      'save_discharge_figs': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # mock the actual output routines
        _delta.make_figure = mock.MagicMock()
        _delta.save_figure = mock.MagicMock()
        _delta.save_grids = mock.MagicMock()

        assert (_delta._save_eta_grids is True)
        assert (_delta._save_metadata is True)

        # check for the netcdf file
        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)
        nc_size_before = os.path.getsize(exp_path_nc)
        assert nc_size_before > 0  # saved once already / inited

        # update a couple times, should increase on each save
        for _t in range(0, 5):
            _delta.save_grids_and_figs()
            _delta._save_iter += 1

        _delta.make_figure.call_count == 5
        _delta.save_figure.call_count == 5
        _delta.save_grids.call_count == 5

    def test_save_all_figures_no_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_eta_figs': True,
                                      'save_discharge_figs': True,
                                      'save_velocity_figs': True,
                                      'save_stage_figs': True,
                                      'save_depth_figs': True,
                                      'save_sedflux_figs': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # mock the actual output routines
        _delta.make_figure = mock.MagicMock()
        _delta.save_figure = mock.MagicMock()
        _delta.save_grids = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert not os.path.isfile(exp_path_nc)

        for _t in range(0, 5):
            _delta.save_grids_and_figs()
            _delta._save_iter += 1
            _delta._time += _delta._dt

        # assertions
        _delta.make_figure.assert_any_call('eta', 0)
        _delta.make_figure.assert_any_call('stage', 0)
        _delta.make_figure.assert_any_call('depth', 0)
        _delta.make_figure.assert_any_call('qw', 0)
        _delta.make_figure.assert_any_call('uw', 0)
        _delta.make_figure.assert_any_call('qs', 0)
        for _i in range(1, 5):
            _delta.make_figure.assert_any_call('eta', _delta._dt * _i)
        _delta.make_figure.call_count == 5 * 6
        _delta.save_figure.call_count == 5 * 6
        _delta.save_grids.call_count == 0

    def test_save_metadata_no_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_metadata': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # mock the actual output routines
        _delta.make_figure = mock.MagicMock()
        _delta.save_figure = mock.MagicMock()
        _delta.save_grids = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 3):
            _delta.save_grids_and_figs()
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert not ('eta' in ds.variables)
        assert ds['meta']['H_SL'].shape[0] == 4  # init + 3
        assert ds['meta']['L0'][:] == 3

    def test_save_metadata_and_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_metadata': True,
                                      'save_eta_grids': True,
                                      'save_velocity_grids': True,
                                      'f_bedload': 0.25})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # mock the actual output routines
        _delta.make_figure = mock.MagicMock()
        _delta.save_figure = mock.MagicMock()
        _delta.save_grids = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 3):
            _delta.save_grids_and_figs()
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert ('eta' in ds.variables)
        assert ('velocity' in ds.variables)
        assert ds['meta']['H_SL'].shape[0] == 4  # init + 3
        assert ds['meta']['L0'][:] == 3
        assert np.all(ds['meta']['f_bedload'][:] == 0.25)

    def test_save_one_grid_metadata_by_default(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_metadata': False,
                                      'save_eta_grids': True,
                                      'C0_percent': 0.2})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # mock the actual output routines
        _delta.make_figure = mock.MagicMock()
        _delta.save_figure = mock.MagicMock()
        _delta.save_grids = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 6):
            _delta.save_grids_and_figs()
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        _arr = ds.variables['eta']
        assert _arr.shape[1] == _delta.eta.shape[0]
        assert _arr.shape[2] == _delta.eta.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too
        assert ds.groups['meta']['H_SL'].shape[0] == _arr.shape[0]
        assert np.all(ds.groups['meta']['C0_percent'][:] == 0.2)
        assert np.all(ds.groups['meta']['f_bedload'][:] == 0.5)


class TestSaveFigure:

    def test_save_figure(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_eta_figs': True})
        _delta = DeltaModel(input_file=p)

        # make a figure
        fig, ax = plt.subplots()
        ax.imshow(np.random.uniform(0, 1, size=(100, 100)))

        # save two figs with different timesteps
        _delta.save_figure(fig, directory=_delta.prefix,
                           filename_root='eta_',
                           timestep=0)

        _delta.save_figure(fig, directory=_delta.prefix,
                           filename_root='eta_',
                           timestep=1)

        # check for output eta file
        exp_path_png0 = os.path.join(tmp_path / 'out_dir', 'eta_00000.png')
        assert os.path.isfile(exp_path_png0)
        exp_path_png1 = os.path.join(tmp_path / 'out_dir', 'eta_00001.png')
        assert os.path.isfile(exp_path_png1)

    def test_save_figure_sequential(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_eta_figs': True,
                                      'save_figs_sequential': False})
        _delta = DeltaModel(input_file=p)

        # make a figure
        fig, ax = plt.subplots()
        ax.imshow(np.random.uniform(0, 1, size=(100, 100)))

        # save two figs with different timesteps
        _delta.save_figure(fig, directory=_delta.prefix,
                           filename_root='eta_',
                           timestep=0)

        _delta.save_figure(fig, directory=_delta.prefix,
                           filename_root='eta_',
                           timestep=1)

        exp_path_png0 = os.path.join(
            tmp_path / 'out_dir', 'eta_00000.png')
        exp_path_png0_latest = os.path.join(
            tmp_path / 'out_dir', 'eta_latest.png')
        assert not os.path.isfile(exp_path_png0)
        assert os.path.isfile(exp_path_png0_latest)


class TestSaveGrids:

    def test_save_eta_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_eta_grids': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 6):
            _delta.save_grids('eta', _delta.eta, _delta._save_iter)
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert (_delta.log_info.call_count == 6)
        _arr = ds.variables['eta']
        assert _arr.shape[1] == _delta.eta.shape[0]
        assert _arr.shape[2] == _delta.eta.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too

    def test_save_depth_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_depth_grids': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 6):
            _delta.save_grids('depth', _delta.depth, _delta._save_iter)
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert (_delta.log_info.call_count == 6)
        _arr = ds.variables['depth']
        assert _arr.shape[1] == _delta.depth.shape[0]
        assert _arr.shape[2] == _delta.depth.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too

    def test_save_velocity_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_velocity_grids': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 6):
            _delta.save_grids('velocity', _delta.uw, _delta._save_iter)
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert (_delta.log_info.call_count == 6)
        _arr = ds.variables['velocity']
        assert _arr.shape[1] == _delta.uw.shape[0]
        assert _arr.shape[2] == _delta.uw.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too

    def test_save_stage_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_stage_grids': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 6):
            _delta.save_grids('stage', _delta.stage, _delta._save_iter)
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert (_delta.log_info.call_count == 6)
        _arr = ds.variables['stage']
        assert _arr.shape[1] == _delta.stage.shape[0]
        assert _arr.shape[2] == _delta.stage.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too

    def test_save_discharge_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_discharge_grids': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 6):
            _delta.save_grids('discharge', _delta.qw, _delta._save_iter)
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert (_delta.log_info.call_count == 6)
        _arr = ds.variables['discharge']
        assert _arr.shape[1] == _delta.qw.shape[0]
        assert _arr.shape[2] == _delta.qw.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too

    def test_save_sedflux_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_sedflux_grids': True})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _t in range(0, 6):
            _delta.save_grids('sedflux', _delta.qs, _delta._save_iter)
            _delta._save_iter += 1

        # close the file and connect
        _delta.output_netcdf.close()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")

        # assertions
        assert (_delta.log_info.call_count == 6)
        _arr = ds.variables['sedflux']
        assert _arr.shape[1] == _delta.qs.shape[0]
        assert _arr.shape[2] == _delta.qs.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too

    def test_save_grids_exception(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml', {'save_dt': 1})
        _delta = DeltaModel(input_file=p)

        # mock the log_info
        _delta.log_info = mock.MagicMock()

        # no netcdf will be created
        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc) is False

        # trying to save a grid will raise an exception
        with pytest.raises(Exception, match=r"DeltaModel' object .*"):
            _delta.save_grids('sedflux', _delta.qs, _delta._save_iter)

        # assert log was called when exception was raised
        assert (_delta.log_info.call_count == 1)



import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import pyDeltaRCM


# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 10
cm = matplotlib.cm.get_cmap('tab10')

# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')

_shp = delta.eta.shape


# manually call only the necessary paths
delta.init_water_iteration()
delta.run_water_iteration()


NPLOT = 5
hdr = NPLOT // 2

# fig, ax = plt.subplots()
fig = plt.figure(constrained_layout=False, figsize=(6, 9))
gs = fig.add_gridspec(
    nrows=NPLOT+hdr, ncols=5,
    left=0.05, right=0.95, top=0.95, bottom=0.05,
    wspace=0.01, hspace=0.2)

hdr_ax = fig.add_subplot(gs[:hdr, :])
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=hdr_ax, grid=False, cmap='cividis', vmin=-3, vmax=0.5)
hdr_ax.set_xticks([])
hdr_ax.set_yticks([])
hdr_ax.set_xlim(0, delta.W // 2)
hdr_ax.set_ylim(delta.L // 2, 0)


def plot_pxpy_weights(pxpy, i):

    px, py = pxpy

    hdr_ax.text(py, px, str(i-hdr),
                fontsize=9, color='white',
                ha='center', va='center')

    __ax_list = []

    def _make_single_axis(i, j, values, **kwargs):
        __ax = fig.add_subplot(gs[i, j])
        __ax.set_xticks([])
        __ax.set_yticks([])
        __ax.imshow(values, **kwargs)
        for _i in range(3):
            for _j in range(3):
                __ax.text(
                    _j, _i, str(np.round(values[_i, _j], 2)),
                    ha='center', va='center', fontsize=8)
        return __ax

    # grab the data from the fields and put it into axis
    _eta_data = delta.eta[px-1:px+2, py-1:py+2]
    _eta_ax = _make_single_axis(
        i, 0, _eta_data, cmap='cividis', vmin=-3, vmax=0.5)
    __ax_list.append(_eta_ax)

    _dpth_data = delta.depth[px-1:px+2, py-1:py+2]
    _dpth_ax = _make_single_axis(
        i, 1, _dpth_data, cmap='Blues', vmin=0, vmax=5)
    __ax_list.append(_dpth_ax)

    _stge_data = delta.stage[px-1:px+2, py-1:py+2]
    _stge_ax = _make_single_axis(
        i, 2, _stge_data, cmap='viridis', vmin=0, vmax=5)
    __ax_list.append(_stge_ax)

    _vel_data = delta.uw[px-1:px+2, py-1:py+2]
    _vel_ax = _make_single_axis(
        i, 3, _vel_data, cmap='plasma', vmin=0, vmax=1.5)
    __ax_list.append(_vel_ax)

    _wght_data = delta.water_weights[px, py, :].reshape((3, 3))
    _wght_ax = _make_single_axis(
        i, 4, _wght_data, cmap='YlGn', vmin=0, vmax=1)
    __ax_list.append(_wght_ax)

    return __ax_list


# define the example location set
_ex_set = [(10, 94), (20, 78), (38, 72), (22, 87), (47, 43)]


# plot the points
ax0 = plot_pxpy_weights(_ex_set[0], hdr+0)
ax1 = plot_pxpy_weights(_ex_set[1], hdr+1)
ax2 = plot_pxpy_weights(_ex_set[2], hdr+2)
ax3 = plot_pxpy_weights(_ex_set[3], hdr+3)
ax4 = plot_pxpy_weights(_ex_set[4], hdr+4)

# labels
for a, axx in enumerate([ax0, ax1, ax2, ax3, ax4]):
    axx[0].set_ylabel(str(a), rotation=0, labelpad=10,
                      ha='center', va='center')

albls = ['bed elevation\n(eta)', 'depth',
         'stage', 'velocity\n(uw)', 'water weights']
for a, (axx, albl) in enumerate(zip(ax4, albls)):
    axx.set_xlabel(albl)

# show
plt.show()



import warnings

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

import pyDeltaRCM
from pyDeltaRCM import water_tools


# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 10
cm = matplotlib.cm.get_cmap('tab10')


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')
_shp = delta.eta.shape


# determine the index to plot at
pidx = 40
Npts = 5
jidx = np.random.randint(0, delta._Np_water, Npts)

# run an interation from the checkpoint
delta.init_water_iteration()
delta.run_water_iteration()

# here, we recreate some steps of each iteration, in order to set up the
#   arrays needed to display the action of _check_for_loops
#
#   1. extract the water walks
current_inds = delta.free_surf_walk_inds[:, pidx-1]
#
#   2. use water weights and random pick to determine d8 direction
water_weights_flat = delta.water_weights.reshape(-1, 9)
new_direction = water_tools._choose_next_directions(
    current_inds, water_weights_flat)
new_direction = new_direction.astype(np.int)
#
#   3. use the new directions for each parcel to determine the new ind for
#   each parcel
new_inds = water_tools._calculate_new_inds(
                current_inds,
                new_direction,
                delta.ravel_walk_flat)
new_inds0 = np.copy(new_inds)

# choose N indices randomly to jump
new_inds[jidx] = delta.free_surf_walk_inds[jidx, pidx-2]

# copy inputs and then run the function to get new outputs
new_inds, looped = water_tools._check_for_loops(
    delta.free_surf_walk_inds[:, :pidx-1], new_inds, pidx + 1, delta.L0,
    delta.CTR, delta.stage - delta.H_SL)


# make the figure
fig, ax = plt.subplots(1, 2, figsize=(10, 4))

delta.show_attribute('eta', ax=ax[0], grid=False, cmap='cividis')
delta.show_attribute('eta', ax=ax[1], grid=False, cmap='cividis')


def _fill_out_an_axis(_ax, _sc=1):
    for j in range(len(jidx)):
        walk = delta.free_surf_walk_inds[jidx[j], :]
        walk = walk[:pidx-2]
        yend, xend = pyDeltaRCM.shared_tools.custom_unravel(
            walk[-1], _shp)
        ynew, xnew = pyDeltaRCM.shared_tools.custom_unravel(
            new_inds[jidx[j]], _shp)
        pyDeltaRCM.debug_tools.plot_line(
            walk, shape=_shp, color=cm(j),
            multiline=True, nozeros=True, lw=1.5*_sc, ax=_ax)
        _ax.plot(xend, yend,
                 marker='o', ms=3*_sc, color=cm(j))
        _ax.plot(xnew, ynew,
                 marker='o', ms=3*_sc, color=cm(j))

    # return the last set
    return yend, xend, ynew, xnew


# fill out both axes with the same info
_fill_out_an_axis(_ax=ax[0])
yend, xend, ynew, xnew = _fill_out_an_axis(_ax=ax[1], _sc=2)

# sub region of the original image
_f = int(delta.W / delta.L)  # ensure x-y scale same
_s = 10
x1, x2, y1, y2 = xend-(_s*_f), xnew+(_s*_f), yend-(_s), ynew+(_s)
ax[1].set_xlim(x1, x2)
ax[1].set_ylim(y2, y1)
ax[1].set_xticks([])
ax[1].set_yticks([])

_r = Rectangle((x1, y2), _s*_f*2, -_s*2, ec='k', fc='none')
ax[0].add_patch(_r)

plt.show()



import warnings

import numpy as np
import matplotlib.pyplot as plt

import pyDeltaRCM

# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')

_shp = delta.eta.shape

# set up axis
fig, ax = plt.subplots()

# fill in axis2
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=ax, grid=False, cmap='cividis')
ax.set_title('intial bed elevation (m)')


plt.tight_layout()
plt.show()



import glob
import os
import netCDF4
import numpy as np

import unittest.mock as mock

from pyDeltaRCM.model import DeltaModel
from .. import utilities


class TestTimingOutputData:

    def test_update_make_record(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_checkpoint': True})
        _delta = DeltaModel(input_file=p)

        # modify the save interval to be twice dt
        _delta._save_dt = 2 * _delta._dt
        _delta._checkpoint_dt = 2 * _delta._dt

        # mock top-level methods, verify call was made to each
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_info = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.save_the_checkpoint = mock.MagicMock()
        # DO NOT mock output_data our output_checkpoint

        # mock the calls inside output_data
        _delta.save_stratigraphy = mock.MagicMock()
        _delta.save_grids_and_figs = mock.MagicMock()

        # run the timestep: t=0
        #   * should call core, but nothing else after init
        _delta.update()

        # assert calls
        assert _delta.solve_water_and_sediment_timestep.call_count == 1
        assert _delta.apply_subsidence.call_count == 1
        assert _delta.finalize_timestep.call_count == 1
        assert _delta.log_model_time.call_count == 1

        # assert times / counters
        assert _delta.time_iter == int(1)
        assert _delta.time == _delta.dt
        assert _delta.save_time_since_data == _delta._dt
        assert _delta.save_iter == int(1)
        assert _delta._save_time_since_checkpoint == _delta._dt

        # run another step
        #   * should call core steps and outputs
        _delta.update()

        # assert calls
        assert _delta.solve_water_and_sediment_timestep.call_count == 2
        assert _delta.apply_subsidence.call_count == 2
        assert _delta.finalize_timestep.call_count == 2
        assert _delta.log_model_time.call_count == 2

        # assert times / counters
        assert _delta.time_iter == int(2)
        assert _delta.time == 2 * _delta.dt
        assert _delta.save_time_since_data == 0
        assert _delta.save_iter == int(2)
        assert _delta._save_time_since_checkpoint == 0

        # run another step
        #   should call core, but nothing else
        _delta.update()

        # assert calls
        assert _delta.solve_water_and_sediment_timestep.call_count == 3
        assert _delta.apply_subsidence.call_count == 3
        assert _delta.finalize_timestep.call_count == 3
        assert _delta.log_model_time.call_count == 3

        # assert times / counters
        assert _delta.time_iter == int(3)
        assert _delta.time == 3 * _delta.dt
        assert _delta.save_time_since_data == _delta._dt
        assert _delta.save_iter == int(2)
        assert _delta._save_time_since_checkpoint == _delta._dt

    def test_update_saving_intervals_on_cycle(self, tmp_path):
        """dt == 300; save_dt == 600"""
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # modify the timestep and save interval to be twice dt
        _delta._dt = 300
        _delta._save_dt = 2 * _delta._dt
        _delta._save_any_grids = True  # override from settings

        # mock top-level methods, verify call was made to each
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()
        # DO NOT mock output_data

        # mock the calls inside output_data
        _delta.save_grids_and_figs = mock.MagicMock()

        _delta.update()  # no new saves, after init

        assert _delta.time_iter == int(1)
        assert _delta.time == _delta.dt
        assert _delta.save_iter == 1
        assert _delta.save_time_since_data == _delta.dt

        _delta.update()  # saves now

        assert _delta.time_iter == int(2)
        assert _delta.time == 2 * _delta.dt
        assert _delta.save_iter == 2
        assert _delta.save_time_since_data == 0

        _delta.update()  # no saves

        assert _delta.time_iter == int(3)
        assert _delta.save_iter == 2
        assert _delta.save_time_since_data == _delta.dt

        # run for a few to bring iters to 10
        for _ in range(7):
            _delta.update()

        assert _delta.time_iter == int(10)
        assert _delta.save_iter == 6

        assert _delta.time == 3000
        assert _delta._is_finalized is False

    def test_update_saving_intervals_short(self, tmp_path):
        """dt == 300; save_dt == 100"""
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # modify the timestep and save interval to be twice dt
        _delta._dt = 300
        _delta._save_dt = 100
        _delta._save_any_grids = True  # override from settings

        # mock top-level methods, verify call was made to each
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()
        # DO NOT mock output_data

        # mock the calls inside output_data
        _delta.save_grids_and_figs = mock.MagicMock()

        _delta.update()  # save on first iteration

        assert _delta.time_iter == int(1)
        assert _delta.time == _delta.dt
        assert _delta.save_grids_and_figs.call_count == 1

        _delta.update()  # save again

        assert _delta.save_time_since_data == 0
        assert _delta.time_iter == int(2)
        assert _delta.time == 2 * _delta.dt
        assert _delta.save_grids_and_figs.call_count == 2

        _delta.update()  # save again

        assert _delta.save_time_since_data == 0
        assert _delta.time_iter == int(3)
        assert _delta.save_grids_and_figs.call_count == 3

        assert _delta.save_iter == 4  # once during init
        assert _delta.time == 900

    def test_update_saving_intervals_offset_long_not_double(self, tmp_path):
        """dt == 300; save_dt == 500"""
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # modify the timestep and save interval to be twice dt
        _delta._dt = 300
        _delta._save_dt = 500
        _delta._save_any_grids = True  # override from settings

        # mock top-level methods, verify call was made to each
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()
        # DO NOT mock output_data

        # mock the calls inside output_data
        _delta.save_grids_and_figs = mock.MagicMock()

        _delta.update()

        assert _delta.time_iter == int(1)
        assert _delta.time == _delta.dt
        assert _delta.save_grids_and_figs.call_count == 0

        _delta.update()

        assert _delta.time_iter == int(2)
        assert _delta.time == 2 * _delta.dt
        assert _delta.save_grids_and_figs.call_count == 1

        for _ in range(10):
            _delta.update()

        assert _delta.time_iter == int(12)
        assert _delta.time == 3600
        assert _delta.save_grids_and_figs.call_count == 6

    def test_update_saving_intervals_offset_long_over_double(self, tmp_path):
        """dt == 300; save_dt == 1000"""
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # modify the timestep and save interval to be twice dt
        _delta._dt = 300
        _delta._save_dt = 1000
        _delta._save_any_grids = True  # override from settings

        # mock top-level methods, verify call was made to each
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()
        # DO NOT mock output_data

        # mock the calls inside output_data
        _delta.save_grids_and_figs = mock.MagicMock()

        _delta.update()

        assert _delta.time_iter == int(1)
        assert _delta.time == _delta.dt

        _delta.update()

        assert _delta.time_iter == int(2)
        assert _delta.time == 2 * _delta.dt
        assert _delta.save_grids_and_figs.call_count == 0

        _delta.update()

        assert _delta.time_iter == int(3)
        assert _delta.time == 3 * _delta.dt
        assert _delta.save_grids_and_figs.call_count == 0

        _delta.update()

        assert _delta.time == 4 * _delta.dt
        assert _delta.save_grids_and_figs.call_count == 1

        _delta.update()

        assert _delta.time == 5 * _delta.dt
        assert _delta.save_grids_and_figs.call_count == 1

        for _ in range(33):
            _delta.update()

        assert _delta.time == 38 * _delta.dt
        assert _delta.save_grids_and_figs.call_count == 9
        assert _delta._is_finalized is False

    def test_finalize_updated(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock the top-level
        _delta.log_info = mock.MagicMock()
        _delta.output_data = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        # modify the save interval
        _t = 5
        _delta._save_dt = _t * _delta._dt
        _delta._checkpoint_dt = _t * _delta._dt

        # run a mock update / save
        _delta._time = _t * _delta._dt
        _delta._save_iter += int(1)
        _delta._save_time_since_data = 0
        _delta._save_time_since_checkpoint = 0

        # run finalize
        _delta.finalize()

        # assert calls
        #   should only hit top-levels
        assert _delta.log_info.call_count == 2
        assert _delta.output_data.call_count == 0
        assert _delta.output_checkpoint.call_count == 0

        assert _delta._is_finalized is True

    def test_save_one_fig_no_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_strata': False,
                                      'save_eta_figs': True})
        _delta = DeltaModel(input_file=p)

        # mock the timestep computations
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        # check one set images created during init
        img_glob = glob.glob(os.path.join(_delta.prefix, '*.png'))
        nc_glob = glob.glob(os.path.join(_delta.prefix, '*.nc'))
        assert len(img_glob) == 1
        assert len(nc_glob) == 0

        # update the delta a few times
        for _ in range(0, 2):
            _delta.update()

        assert _delta.time_iter == 2.0
        _delta.solve_water_and_sediment_timestep.call_count == 2

        # check for output eta files
        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        exp_path_png0 = os.path.join(tmp_path / 'out_dir', 'eta_00000.png')
        exp_path_png1 = os.path.join(tmp_path / 'out_dir', 'eta_00001.png')
        exp_path_png2 = os.path.join(tmp_path / 'out_dir', 'eta_00002.png')
        exp_path_png3 = os.path.join(tmp_path / 'out_dir', 'eta_00003.png')
        assert not os.path.isfile(exp_path_nc)
        assert os.path.isfile(exp_path_png0)
        assert os.path.isfile(exp_path_png1)
        assert os.path.isfile(exp_path_png2)
        assert not os.path.isfile(exp_path_png3)

    def test_save_one_fig_one_grid(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_eta_grids': True,
                                      'save_discharge_figs': True})
        _delta = DeltaModel(input_file=p)

        # mock the timestep computations
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)
        nc_size_before = os.path.getsize(exp_path_nc)
        assert nc_size_before > 0

        # update a couple times, should increase on each save
        for _ in range(0, 2):
            _delta.update()

        nc_size_middle = os.path.getsize(exp_path_nc)
        assert _delta.time_iter == 2.0
        assert nc_size_middle > nc_size_before

        # now finalize, and file size should stay the same
        _delta.finalize()
        nc_size_after = os.path.getsize(exp_path_nc)
        assert _delta.time_iter == 2.0
        assert nc_size_after == nc_size_middle
        assert nc_size_after > nc_size_before

    def test_save_metadata_no_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_metadata': True})
        _delta = DeltaModel(input_file=p)

        # mock the timestep computations
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _ in range(0, 2):
            _delta.update()

        assert _delta.time_iter == 2.0
        _delta.solve_water_and_sediment_timestep.call_count == 2

        _delta.finalize()

        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")
        assert not ('eta' in ds.variables)
        assert ds['meta']['H_SL'].shape[0] == 3
        assert ds['meta']['L0'][:] == 3

    def test_save_subsidence_metadata_no_grids(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'toggle_subsidence': True,
                                      'start_subsidence': 0,
                                      'subsidence_rate': 1,
                                      'save_metadata': True})
        _delta = DeltaModel(input_file=p)

        # mock the timestep computations
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _ in range(0, 2):
            _delta.update()

        assert _delta.time_iter == 2.0
        _delta.solve_water_and_sediment_timestep.call_count == 2

        _delta.finalize()

        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")
        assert not ('eta' in ds.variables)
        assert ds['meta']['H_SL'].shape[0] == 3
        assert ds['meta']['L0'][:] == 3
        assert ds['meta']['sigma'].shape == _delta.sigma.shape
        assert np.all(ds['meta']['sigma'] == _delta.sigma)
        assert ds['meta']['start_subsidence'][:] == 0

    def test_save_one_grid_metadata_by_default(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_dt': 1,
                                      'save_metadata': False,
                                      'save_eta_grids': True,
                                      'C0_percent': 0.2})
        _delta = DeltaModel(input_file=p)

        # mock the timestep computations
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        exp_path_nc = os.path.join(
            tmp_path / 'out_dir', 'pyDeltaRCM_output.nc')
        assert os.path.isfile(exp_path_nc)

        for _ in range(0, 6):
            _delta.update()

        assert _delta.time_iter == 6.0
        _delta.solve_water_and_sediment_timestep.call_count == 6

        _delta.finalize()
        ds = netCDF4.Dataset(exp_path_nc, "r", format="NETCDF4")
        _arr = ds.variables['eta']
        assert _arr.shape[1] == _delta.eta.shape[0]
        assert _arr.shape[2] == _delta.eta.shape[1]
        assert ('meta' in ds.groups)  # if any grids, save meta too
        assert ds.groups['meta']['H_SL'].shape[0] == _arr.shape[0]
        assert np.all(ds.groups['meta']['C0_percent'][:] == 0.2)
        assert np.all(ds.groups['meta']['f_bedload'][:] == 0.5)



import pytest
import os
import platform
import locale
import subprocess

from .. import utilities
import pyDeltaRCM as _pyimportedalias


class TestCommandLineInterfaceDirectly:

    # test the entry points
    def test_entry_point_installed_call(self, tmp_path):
        """
        test calling the command line feature with a config file.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'timesteps', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'save_dt', 300)
        utilities.write_parameter_to_file(f, 'save_eta_figs', True)
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        f.close()
        subprocess.check_output(['pyDeltaRCM',
                                 '--config', str(p)])
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        exp_path_png0 = os.path.join(tmp_path / 'test', 'eta_00000.png')
        exp_path_png1 = os.path.join(tmp_path / 'test', 'eta_00001.png')
        assert os.path.isfile(exp_path_nc)
        assert os.path.isfile(exp_path_png0)
        assert os.path.isfile(exp_path_png1)

    def test_entry_point_python_main_call(self, tmp_path):
        """
        test calling the python hook command line feature with a config file.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'timesteps', 1)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'save_dt', 300)
        utilities.write_parameter_to_file(f, 'save_eta_figs', True)
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        f.close()
        subprocess.check_output(['python', '-m', 'pyDeltaRCM',
                                 '--config', str(p)])
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        exp_path_png = os.path.join(tmp_path / 'test', 'eta_00000.png')
        exp_path_png1 = os.path.join(tmp_path / 'test', 'eta_00001.png')
        exp_path_png2 = os.path.join(tmp_path / 'test', 'eta_00002.png')
        assert os.path.isfile(exp_path_nc)
        assert os.path.isfile(exp_path_png)
        assert os.path.isfile(exp_path_png1)
        assert not os.path.isfile(exp_path_png2)

    def test_entry_point_python_main_call_dryrun(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'timesteps', 1)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        subprocess.check_output(['python', '-m', 'pyDeltaRCM',
                                 '--config', str(p),
                                 '--dryrun'])
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        exp_path_png = os.path.join(tmp_path / 'test', 'eta_00000.png')
        assert not os.path.isfile(exp_path_nc)   # does not exist because --dryrun
        assert not os.path.isfile(exp_path_png)  # does not exist because --dryrun

    def test_entry_point_python_main_call_timesteps(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'save_dt', 300)
        utilities.write_parameter_to_file(f, 'save_eta_figs', True)
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        f.close()
        subprocess.check_output(['python', '-m', 'pyDeltaRCM',
                                 '--config', str(p),
                                 '--timesteps', '2'])
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        exp_path_png = os.path.join(tmp_path / 'test', 'eta_00000.png')
        assert os.path.isfile(exp_path_nc)
        assert os.path.isfile(exp_path_png)

    def test_error_if_no_timesteps(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_output(['python', '-m', 'pyDeltaRCM',
                                     '--config', str(p)])

    def test_entry_point_timesteps(self, tmp_path):
        """
        test calling the command line feature with a config file.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'save_dt', 300)
        utilities.write_parameter_to_file(f, 'save_eta_figs', True)
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        f.close()
        subprocess.check_output(['pyDeltaRCM',
                                 '--config', str(p), '--timesteps', '2'])
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        exp_path_png0 = os.path.join(tmp_path / 'test', 'eta_00000.png')
        exp_path_png1 = os.path.join(tmp_path / 'test', 'eta_00001.png')
        assert os.path.isfile(exp_path_nc)
        assert os.path.isfile(exp_path_png0)
        assert os.path.isfile(exp_path_png1)

    def test_entry_point_time(self, tmp_path):
        """
        test calling the command line feature with a config file.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'Length', 10.0)
        utilities.write_parameter_to_file(f, 'Width', 10.0)
        utilities.write_parameter_to_file(f, 'seed', 0)
        utilities.write_parameter_to_file(f, 'dx', 1.0)
        utilities.write_parameter_to_file(f, 'L0_meters', 1.0)
        utilities.write_parameter_to_file(f, 'N0_meters', 2.0)
        utilities.write_parameter_to_file(f, 'Np_water', 10)
        utilities.write_parameter_to_file(f, 'Np_sed', 10)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'save_dt', 300)
        utilities.write_parameter_to_file(f, 'save_eta_figs', True)
        utilities.write_parameter_to_file(f, 'save_eta_grids', True)
        f.close()
        subprocess.check_output(['pyDeltaRCM',
                                 '--config', str(p), '--time', '1000'])
        exp_path_nc = os.path.join(tmp_path / 'test', 'pyDeltaRCM_output.nc')
        exp_path_png0 = os.path.join(tmp_path / 'test', 'eta_00000.png')
        exp_path_png1 = os.path.join(tmp_path / 'test', 'eta_00001.png')
        assert os.path.isfile(exp_path_nc)
        assert os.path.isfile(exp_path_png0)
        assert os.path.isfile(exp_path_png1)

    def test_version_call(self):
        """
        test calling the command line feature to query the version.
        """
        encoding = locale.getpreferredencoding()
        printed1 = subprocess.run(['pyDeltaRCM', '--version'],
                                  stdout=subprocess.PIPE, encoding=encoding)
        _exp_str1 = 'pyDeltaRCM ' + _pyimportedalias.__version__ + '\n'
        assert printed1.stdout == _exp_str1
        printed2 = subprocess.run(
            ['python', '-m', 'pyDeltaRCM', '--version'],
            stdout=subprocess.PIPE, encoding=encoding)
        _exp_str2 = 'pyDeltaRCM ' + _pyimportedalias.__version__ + '\n'
        assert printed2.stdout == _exp_str2
